"""List suitable replication datacenters for the given volume."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import columns as column_helper
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting

COLUMNS = [
    column_helper.Column('ID', ('id',), mask="id"),
    column_helper.Column('Long Name', ('longName',), mask="longName"),
    column_helper.Column('Short Name', ('name',), mask="name"),
]

DEFAULT_COLUMNS = [
    'ID',
    'Long Name',
    'Short Name',
]


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('volume-id')
@click.option('--columns',
              callback=column_helper.get_formatter(COLUMNS),
              help=f"Columns to display. Options: {', '.join(column.name for column in COLUMNS)}",
              default=','.join(DEFAULT_COLUMNS))
@click.option('--sortby', help='Column to sort by', default='Long Name')
@environment.pass_env
def cli(env, columns, sortby, volume_id):
    """List suitable replication datacenters for the given volume."""
    block_storage_manager = SoftLayer.BlockStorageManager(env.client)

    legal_centers = block_storage_manager.get_replication_locations(
        volume_id
    )

    if not legal_centers:
        click.echo("No data centers compatible for replication.")
    else:
        table = formatting.KeyValueTable(columns.columns)
        table.sortby = sortby
        for legal_center in legal_centers:
            table.add_row([value or formatting.blank()
                           for value in columns.row(legal_center)])

        env.fout(table)



"""
    SoftLayer.managers
    ~~~~~~~~~~~~~~~~~~
    Managers mask out a lot of the complexities of using the API into classes
    that provide a simpler interface to various services. These are
    higher-level interfaces to the SoftLayer API.

    :license: MIT, see LICENSE for more details.
"""
from SoftLayer.managers.account import AccountManager
from SoftLayer.managers.bandwidth import BandwidthManager
from SoftLayer.managers.block import BlockStorageManager
from SoftLayer.managers.dedicated_host import DedicatedHostManager
from SoftLayer.managers.dns import DNSManager
from SoftLayer.managers.event_log import EventLogManager
from SoftLayer.managers.file import FileStorageManager
from SoftLayer.managers.firewall import FirewallManager
from SoftLayer.managers.hardware import HardwareManager
from SoftLayer.managers.image import ImageManager
from SoftLayer.managers.license import LicensesManager
from SoftLayer.managers.load_balancer import LoadBalancerManager
from SoftLayer.managers.metadata import MetadataManager
from SoftLayer.managers.network import NetworkManager
from SoftLayer.managers.object_storage import ObjectStorageManager
from SoftLayer.managers.ordering import OrderingManager
from SoftLayer.managers.search import SearchManager
from SoftLayer.managers.sshkey import SshKeyManager
from SoftLayer.managers.ssl import SSLManager
from SoftLayer.managers.tags import TagManager
from SoftLayer.managers.ticket import TicketManager
from SoftLayer.managers.user import UserManager
from SoftLayer.managers.vs import VSManager
from SoftLayer.managers.vs_capacity import CapacityManager
from SoftLayer.managers.vs_placement import PlacementManager

__all__ = [
    'AccountManager',
    'BandwidthManager',
    'BlockStorageManager',
    'CapacityManager',
    'DedicatedHostManager',
    'DNSManager',
    'EventLogManager',
    'FileStorageManager',
    'FirewallManager',
    'HardwareManager',
    'ImageManager',
    'LicensesManager',
    'LoadBalancerManager',
    'MetadataManager',
    'NetworkManager',
    'ObjectStorageManager',
    'OrderingManager',
    'PlacementManager',
    'SearchManager',
    'SshKeyManager',
    'SSLManager',
    'TagManager',
    'TicketManager',
    'UserManager',
    'VSManager',
]



"""List NAS accounts."""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer import utils


@click.command(cls=SLCommand)
@environment.pass_env
def cli(env):
    """List NAS accounts."""

    account = env.client['Account']

    nas_accounts = account.getNasNetworkStorage(
        mask='eventCount,serviceResource[datacenter.name]')

    table = formatting.Table(['id', 'datacenter', 'size', 'server'])

    for nas_account in nas_accounts:
        table.add_row([
            nas_account['id'],
            utils.lookup(nas_account,
                         'serviceResource',
                         'datacenter',
                         'name') or formatting.blank(),
            formatting.FormattedItem(
                nas_account.get('capacityGb', formatting.blank()),
                "%dGB" % nas_account.get('capacityGb', 0)),
            nas_account.get('serviceResourceBackendIpAddress',
                            formatting.blank())])

    env.fout(table)



getObject = {
    'accountId': 1234,
    'id': 1234,
    'name': 'TestQuote1234',
    'quoteKey': '1234test4321',
    'order': {
        'id': 37623333,
        'items': [
            {
                'categoryCode': 'guest_core',
                'description': '4 x 2.0 GHz or higher Cores',
                'id': 468394713,
                'itemId': 859,
                'itemPriceId': '1642',
                'oneTimeAfterTaxAmount': '0',
                'oneTimeFee': '0',
                'oneTimeFeeTaxRate': '0',
                'oneTimeTaxAmount': '0',
                'quantity': 1,
                'recurringAfterTaxAmount': '0',
                'recurringFee': '0',
                'recurringTaxAmount': '0',
                'setupAfterTaxAmount': '0',
                'setupFee': '0',
                'setupFeeDeferralMonths': None,
                'setupFeeTaxRate': '0',
                'setupTaxAmount': '0',
                'package': {'id': 46, 'keyName': 'CLOUD_SERVER'}
            },
        ]
    }
}

getRecalculatedOrderContainer = {
    'presetId': '',
    'prices': [{
        'id': 1921
    }],
    'quantity': 1,
    'packageId': 50,
    'useHourlyPricing': '',
    'reservedCapacityId': '',

}

verifyOrder = {
    'orderId': 1234,
    'orderDate': '2013-08-01 15:23:45',
    'useHourlyPricing': False,
    'prices': [{
        'id': 1,
        'laborFee': '2',
        'oneTimeFee': '2',
        'oneTimeFeeTax': '.1',
        'quantity': 1,
        'recurringFee': '2',
        'recurringFeeTax': '.1',
        'hourlyRecurringFee': '2',
        'setupFee': '1',
        'item': {'id': 1, 'description': 'this is a thing', 'keyName': 'TheThing'},
    }]}

placeOrder = {
    'orderId': 1234,
    'orderDate': '2013-08-01 15:23:45',
    'orderDetails': {
        'prices': [{
            'id': 1,
            'laborFee': '2',
            'oneTimeFee': '2',
            'oneTimeFeeTax': '.1',
            'quantity': 1,
            'recurringFee': '2',
            'recurringFeeTax': '.1',
            'hourlyRecurringFee': '2',
            'setupFee': '1',
            'item': {'id': 1, 'description': 'this is a thing'},
        }],
        'virtualGuests': [{
            'id': 1234567,
            'globalIdentifier': '1a2b3c-1701',
            'fullyQualifiedDomainName': 'test.guest.com'
        }],
    },
    'placedOrder': {
        'id': 37985543,
        'orderQuoteId': 2639077,
        'orderTypeId': 4,
        'status': 'PENDING_AUTO_APPROVAL',
        'items': [
            {
                'categoryCode': 'guest_core',
                'description': '4 x 2.0 GHz or higher Cores',
                'id': 472527133,
                'itemId': 859,
                'itemPriceId': '1642',
                'laborFee': '0',
                'oneTimeFee': '0',
                'recurringFee': '0',
                'setupFee': '0',
            }
        ]
    }
}

saveQuote = getObject
deleteQuote = getObject



"""Detail firewall."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import firewall
from SoftLayer.CLI import formatting
from SoftLayer import utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--password', is_flag=True,
              help="Display FortiGate username and FortiGate password to multi vlans.")
@environment.pass_env
def cli(env, identifier, password):
    """Detail firewall.

    EXAMPLES:

        slcli firewall detail vs:12345

        slcli firewall detail --credentials true multiVlan:456789
    """

    mgr = SoftLayer.FirewallManager(env.client)

    firewall_type, firewall_id = firewall.parse_id(identifier)

    if firewall_type in ('vs', 'server', 'vlan', 'multiVlan'):

        if firewall_type == 'vlan':
            _firewall = mgr.get_instance(firewall_id)

            table = formatting.KeyValueTable(['name', 'value'])
            table.align['name'] = 'r'
            table.align['value'] = 'l'

            table.add_row(['id', _firewall.get('id')])
            table.add_row(['primaryIpAddress', _firewall.get('primaryIpAddress')])
            table.add_row(['datacenter', utils.lookup(_firewall, 'datacenter', 'longName')])
            table.add_row(['networkVlan', utils.lookup(_firewall, 'networkVlan', 'name')])
            table.add_row(['networkVlaniD', utils.lookup(_firewall, 'networkVlan', 'id')])

            rules = mgr.get_dedicated_fwl_rules(firewall_id)
            table.add_row(['rules', get_rules_table(rules)])

        if firewall_type == 'multiVlan':
            _firewall = mgr.get_instance(firewall_id)

            table = formatting.KeyValueTable(['name', 'value'])
            table.align['name'] = 'r'
            table.align['value'] = 'l'

            table.add_row(['name', utils.lookup(_firewall, 'networkGateway', 'name')])
            table.add_row(['datacenter', utils.lookup(_firewall, 'datacenter', 'longName')])
            table.add_row(['public ip', utils.lookup(_firewall, 'networkGateway', 'publicIpAddress', 'ipAddress')])
            table.add_row(['private ip', utils.lookup(_firewall, 'networkGateway', 'privateIpAddress', 'ipAddress')])
            table.add_row(['public  ipv6', utils.lookup(_firewall, 'networkGateway', 'publicIpv6Address', 'ipAddress')])
            table.add_row(['public vlan', utils.lookup(_firewall, 'networkGateway', 'publicVlan', 'vlanNumber')])
            table.add_row(['private vlan', utils.lookup(_firewall, 'networkGateway', 'privateVlan', 'vlanNumber')])
            table.add_row(['type', _firewall.get('firewallType')])

            if password:
                table.add_row(['fortiGate username', utils.lookup(_firewall, 'managementCredentials', 'username')])
                table.add_row(['fortiGate password', utils.lookup(_firewall, 'managementCredentials', 'password')])

            rules = mgr.get_dedicated_fwl_rules(firewall_id)
            if len(rules) != 0:
                table.add_row(['rules', get_rules_table(rules)])
            else:
                table.add_row(['rules', '-'])

        if firewall_type == 'vs' or firewall_type == 'server':
            rules = mgr.get_standard_fwl_rules(firewall_id)
            table = get_rules_table(rules)

        env.fout(table)

    else:
        click.secho('Invalid firewall type %s: firewall type should be either vlan, multiVlan, vs or server.'
                    % firewall_type, fg='red')


def get_rules_table(rules):
    """Helper to format the rules into a table.

    :param list rules: A list containing the rules of the firewall
    :returns: a formatted table of the firewall rules
    """
    table = formatting.Table(['#', 'action', 'protocol', 'src_ip', 'src_mask',
                              'dest', 'dest_mask'])
    table.sortby = '#'
    for rule in rules:
        table.add_row([
            rule['orderValue'],
            rule['action'],
            rule['protocol'],
            rule['sourceIpAddress'],
            utils.lookup(rule, 'sourceIpSubnetMask'),
            '%s:%s-%s' % (rule['destinationIpAddress'],
                          rule['destinationPortRangeStart'],
                          rule['destinationPortRangeEnd']),
            utils.lookup(rule, 'destinationIpSubnetMask')])
    return table



"""Import an image."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('name')
@click.argument('uri')
@click.option('--note',
              default="",
              help="The note to be applied to the imported template")
@click.option('--os-code',
              help="The referenceCode of the operating system software"
                   " description for the imported VHD, ISO, or RAW image")
@click.option('--ibm-api-key',
              default=None,
              help="The IBM Cloud API Key with access to IBM Cloud Object "
                   "Storage instance and IBM KeyProtect instance. For help "
                   "creating this key see https://cloud.ibm.com/docs/"
                   "cloud-object-storage?topic=cloud-object-storage"
                   "-iam-overview#iam-overview-service-id-api-key")
@click.option('--root-key-crn',
              default=None,
              help="CRN of the root key in your KMS instance")
@click.option('--wrapped-dek',
              default=None,
              help="Wrapped Data Encryption Key provided by IBM KeyProtect. "
                   "For more info see "
                   "https://cloud.ibm.com/docs/key-protect?topic=key-protect-wrap-keys")
@click.option('--cloud-init',
              is_flag=True,
              help="Specifies if image is cloud-init")
@click.option('--byol',
              is_flag=True,
              help="Specifies if image is bring your own license")
@click.option('--is-encrypted',
              is_flag=True,
              help="Specifies if image is encrypted")
@environment.pass_env
def cli(env, name, note, os_code, uri, ibm_api_key, root_key_crn, wrapped_dek,
        cloud_init, byol, is_encrypted):
    """Import an image.

    The URI for an object storage object (.vhd/.iso file) of the format:
    swift://<objectStorageAccount>@<cluster>/<container>/<objectPath>
    or cos://<regionName>/<bucketName>/<objectPath> if using IBM Cloud
    Object Storage
    """

    image_mgr = SoftLayer.ImageManager(env.client)
    result = image_mgr.import_image_from_uri(
        name=name,
        note=note,
        os_code=os_code,
        uri=uri,
        ibm_api_key=ibm_api_key,
        root_key_crn=root_key_crn,
        wrapped_dek=wrapped_dek,
        cloud_init=cloud_init,
        byol=byol,
        is_encrypted=is_encrypted
    )

    if not result:
        raise exceptions.CLIAbort("Failed to import Image")

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'
    table.add_row(['name', result['name']])
    table.add_row(['id', result['id']])
    table.add_row(['created', result['createDate']])
    table.add_row(['guid', result['globalIdentifier']])
    env.fout(table)



getInvoiceTopLevelItems = [
    {
        "categoryCode": "sov_sec_ip_addresses_priv",
        "createDate": "2018-04-04T23:15:20-06:00",
        "description": "64 Portable Private IP Addresses",
        "id": 724951323,
        "oneTimeAfterTaxAmount": "0",
        "recurringAfterTaxAmount": "0",
        "hostName": "bleg",
        "domainName": "beh.com",
        "category": {"name": "Private (only) Secondary VLAN IP Addresses"},
        "children": [
            {
                "id": 12345,
                "category": {"name": "Fake Child Category"},
                "description": "Blah",
                "oneTimeAfterTaxAmount": 55.50,
                "recurringAfterTaxAmount": 0.10
            }
        ],
        "location": {"name": "fra02"}
    },
    {
        "categoryCode": "reserved_capacity",
        "createDate": "2024-07-03T22:08:36-07:00",
        "description": "B1.1x2 (1 Year Term) (721hrs * .025)",
        "id": 1111222,
        "oneTimeAfterTaxAmount": "0",
        "recurringAfterTaxAmount": "18.03",
        "category": {"name": "Reserved Capacity"},
        "children": [
            {
                "description": "1 x 2.0 GHz or higher Core",
                "id": 29819,
                "oneTimeAfterTaxAmount": "0",
                "recurringAfterTaxAmount": "10.00",
                "category": {"name": "Computing Instance"}
            },
            {
                "description": "2 GB",
                "id": 123456,
                "oneTimeAfterTaxAmount": "0",
                "recurringAfterTaxAmount": "2.33",
                "category": {"name": "RAM"}
            }
        ],
        "location": {"name": "dal10"}
    }
]



"""
    SoftLayer.consts
    ~~~~~~~~~~~~~~~~
    Contains constants used throughout the library

    :license: MIT, see LICENSE for more details.
"""
VERSION = 'v6.2.8'
API_PUBLIC_ENDPOINT = 'https://api.softlayer.com/xmlrpc/v3.1/'
API_PRIVATE_ENDPOINT = 'https://api.service.softlayer.com/xmlrpc/v3.1/'
API_PUBLIC_ENDPOINT_REST = 'https://api.softlayer.com/rest/v3.1/'
API_PRIVATE_ENDPOINT_REST = 'https://api.service.softlayer.com/rest/v3.1/'
USER_AGENT = "softlayer-python/%s" % VERSION
CONFIG_FILE = "~/.softlayer"



"""Create security groups."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--name', '-n',
              help="The name of the security group")
@click.option('--description', '-d',
              help="The description of the security group")
@environment.pass_env
def cli(env, name, description):
    """Create a security group."""
    mgr = SoftLayer.NetworkManager(env.client)

    result = mgr.create_securitygroup(name, description)
    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'
    table.add_row(['id', result['id']])
    table.add_row(['name',
                   result.get('name') or formatting.blank()])
    table.add_row(['description',
                   result.get('description') or formatting.blank()])
    table.add_row(['created', result['createDate']])

    env.fout(table)



"""Setup CLI configuration."""
# :license: MIT, see LICENSE for more details.
import webbrowser

import configparser
import json
import os.path
import requests

import click

import SoftLayer
from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import config
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.consts import USER_AGENT
from SoftLayer import utils


def get_api_key(client, username, secret):  # pylint: disable=inconsistent-return-statements
    """Attempts API-Key and password auth to get an API key.

    This will also generate an API key if one doesn't exist
    """

    # Try to use a client with username/api key
    if len(secret) == 64 or username == 'apikey':
        try:
            client['Account'].getCurrentUser()
            return secret
        except SoftLayer.SoftLayerAPIError as ex:
            if 'invalid api token' not in ex.faultString.lower():
                raise
    else:
        if isinstance(client, SoftLayer.API.IAMClient):
            client.authenticate_with_iam_token(secret)
        else:
            # Try to use a client with username/password
            client.authenticate_with_password(username, secret)

        user_record = client.call('Account', 'getCurrentUser', mask='id, apiAuthenticationKeys')
        api_keys = user_record['apiAuthenticationKeys']
        if len(api_keys) == 0:
            return client.call('User_Customer', 'addApiAuthenticationKey', id=user_record['id'])
        return api_keys[0]['authenticationKey']


@click.command(cls=SLCommand)
@click.option('-a', '--auth', type=click.Choice(['ibmid', 'cloud_key', 'classic_key', 'sso']),
              help="Select a method of authentication.", default='classic_key', show_default=True)
@environment.pass_env
def cli(env, auth):
    """Setup the ~/.softlayer file with username and apikey.

    [Auth Types]

    ibmid: Requires your cloud.ibm.com username and password, and generates a classic infrastructure API key.

    cloud_key: A 32 character API key. Username will be 'apikey'

    classic_key: A 64 character API key used in the Softlayer/Classic Infrastructure systems.

    sso: For users with @ibm.com email addresses.
    """
    username = None
    api_key = None

    timeout = 0
    defaults = config.get_settings_from_client(env.client, env.theme)
    endpoint_url = get_endpoint_url(env, defaults.get('endpoint_url', 'public'))
    # Get ths username and API key
    if auth == 'ibmid':
        username, api_key = ibmid_login(env)

    elif auth == 'cloud_key':
        username = 'apikey'
        secret = env.getpass('Classic Infrastructure API Key', default=defaults['api_key'])
        new_client = SoftLayer.Client(username=username, api_key=secret, endpoint_url=endpoint_url, timeout=timeout)
        env.client = new_client
        env.client.transport = SoftLayer.DebugTransport(new_client.transport)
        api_key = get_api_key(new_client, username, secret)

    elif auth == 'sso':
        username, api_key = sso_login(env)

    else:
        username = env.input('Classic Infrastructure Username', default=defaults['username'])
        secret = env.getpass('Classic Infrastructure API Key', default=defaults['api_key'])
        new_client = SoftLayer.Client(username=username, api_key=secret, endpoint_url=endpoint_url, timeout=timeout)
        env.client = new_client
        env.client.transport = SoftLayer.DebugTransport(new_client.transport)
        api_key = get_api_key(new_client, username, secret)

    # Ask for timeout, convert to float, then to int
    timeout = int(float(env.input('Timeout', default=defaults['timeout'] or 0)))

    # Ask theme for console
    theme = env.input('Theme (dark/light)', default=defaults['theme'] or 'dark')

    path = '~/.softlayer'
    if env.config_file:
        path = env.config_file
    config_path = os.path.expanduser(path)

    env.out(env.fmt(config.config_table({'username': username,
                                         'api_key': api_key,
                                         'endpoint_url': endpoint_url,
                                         'timeout': timeout,
                                         'theme': theme})))

    if not formatting.confirm('Are you sure you want to write settings to "%s"?' % config_path, default=True):
        raise exceptions.CLIAbort('Aborted.')

    # Persist the config file. Read the target config file in before
    # setting the values to avoid clobbering settings
    parsed_config = configparser.RawConfigParser()
    parsed_config.read(config_path)
    try:
        parsed_config.add_section('softlayer')
    except configparser.DuplicateSectionError:
        pass

    parsed_config.set('softlayer', 'username', username)
    parsed_config.set('softlayer', 'api_key', api_key)
    parsed_config.set('softlayer', 'endpoint_url', endpoint_url)
    parsed_config.set('softlayer', 'timeout', timeout)
    parsed_config.set('softlayer', 'theme', theme)

    config_fd = os.fdopen(os.open(config_path, (os.O_WRONLY | os.O_CREAT | os.O_TRUNC), 0o600), 'w')
    try:
        parsed_config.write(config_fd)
    finally:
        config_fd.close()

    env.fout("Configuration Updated Successfully")


def get_endpoint_url(env, endpoint='public'):
    """Gets the Endpoint to use."""

    endpoint_type = env.input('Endpoint (public|private|custom)', default=endpoint)
    endpoint_type = endpoint_type.lower()

    if endpoint_type == 'public':
        endpoint_url = SoftLayer.API_PUBLIC_ENDPOINT
    elif endpoint_type == 'private':
        endpoint_url = SoftLayer.API_PRIVATE_ENDPOINT
    else:
        if endpoint_type == 'custom':
            endpoint_url = env.input('Endpoint URL', default=endpoint)
        else:
            endpoint_url = endpoint_type
    return endpoint_url


def ibmid_login(env):
    """Uses an IBMid and Password to get an access token, and that access token to get an API key"""
    email = env.input("Email").strip()
    password = env.getpass("Password").strip()

    client = SoftLayer.API.IAMClient(config_file=env.config_file)

    # STEP 1: Get the base IAM Token with a username/password
    tokens = client.authenticate_with_password(email, password)

    # STEP 2: Figure out which account we want to use
    account = get_accounts(env, tokens['access_token'])

    # STEP 3: Refresh Token, using a specific account this time.
    tokens = client.refresh_iam_token(tokens['refresh_token'], account['account_id'], account['ims_id'])

    # STEP 4: Get or create the Classic Infrastructure API key
    user = client.call('SoftLayer_Account', 'getCurrentUser', mask="mask[id,username,apiAuthenticationKeys]")

    if len(user.get('apiAuthenticationKeys', [])) == 0:
        env.fout(f"Creating a Classic Infrastrucutre API key for {user['username']}")
        api_key = client.call('User_Customer', 'addApiAuthenticationKey', id=user['id'])
    else:
        api_key = user['apiAuthenticationKeys'][0]['authenticationKey']

    return user.get('username'), api_key


def get_accounts(env, a_token):
    """Gets account list from accounts.cloud.ibm.com/v1/accounts"""
    iam_client = requests.Session()

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'User-Agent': USER_AGENT,
        'Accept': 'application/json'
    }
    headers['Authorization'] = f'Bearer {a_token}'
    response = iam_client.request(
        'GET',
        'https://accounts.cloud.ibm.com/v1/accounts',
        headers=headers
    )

    response.raise_for_status()

    accounts = json.loads(response.text)
    selected = None
    ims_id = None
    if accounts.get('total_results', 0) == 1:
        selected = accounts['resources'][0]
    else:
        env.fout("Select an Account...")
        counter = 1
        for selected in accounts.get('resources', []):
            links = utils.lookup(selected, 'metadata', 'linked_accounts') or []
            for link in links:
                if link.get('origin') == "IMS":
                    ims_id = link.get('id')
            if ims_id is None:
                ims_id = "Unlinked"
            env.fout(f"{counter}: {utils.lookup(selected, 'entity', 'name')} ({ims_id})")
            counter = counter + 1
        ims_id = None  # Reset ims_id to avoid any mix-match or something.
        choice = click.prompt('Enter a number', type=int)
        # Test to make sure choice is not out of bounds...
        selected = accounts['resources'][choice - 1]

    account_id = utils.lookup(selected, 'metadata', 'guid')
    links = utils.lookup(selected, 'metadata', 'linked_accounts') or []
    for link in links:
        if link.get('origin') == "IMS":
            ims_id = link.get('id')

    print(f"Using account {utils.lookup(selected, 'entity', 'name')}")
    return {"account_id": account_id, "ims_id": ims_id}


def get_sso_url():
    """Gets the URL for using SSO Tokens"""

    iam_client = requests.Session()

    headers = {
        'Content-Type': 'application/json',
        'User-Agent': USER_AGENT,
        'Accept': 'application/json'
    }
    response = iam_client.request(
        'GET',
        'https://iam.cloud.ibm.com/identity/.well-known/openid-configuration',
        headers=headers
    )

    response.raise_for_status()
    data = json.loads(response.text)
    return data.get('passcode_endpoint')


def sso_login(env):
    """Uses a SSO token to get a SL apikey"""
    passcode_url = get_sso_url()
    env.fout(f"Get a one-time code from {passcode_url} to proceed.")
    open_browser = env.input("Open the URL in the default browser? [Y/n]", default='Y')
    if open_browser.lower() == 'y':
        webbrowser.open(passcode_url)
    passcode = env.input("One-time code")
    client = SoftLayer.API.IAMClient(config_file=env.config_file)

    # STEP 1: Get the base IAM Token with a username/password
    tokens = client.authenticate_with_passcode(passcode)

    # STEP 2: Figure out which account we want to use
    account = get_accounts(env, tokens['access_token'])

    # STEP 3: Refresh Token, using a specific account this time.
    tokens = client.refresh_iam_token(tokens['refresh_token'], account['account_id'], account['ims_id'])

    # STEP 4: Get or create the Classic Infrastructure API key
    # client.authenticate_with_iam_token(tokens['access_token'])
    user = client.call('SoftLayer_Account', 'getCurrentUser', mask="mask[id,username,apiAuthenticationKeys]")

    if len(user.get('apiAuthenticationKeys', [])) == 0:
        env.fout(f"Creating a Classic Infrastrucutre API key for {user['username']}")
        api_key = client.call('User_Customer', 'addApiAuthenticationKey', id=user['id'])
    else:
        api_key = user['apiAuthenticationKeys'][0]['authenticationKey']
    return user.get('username'), api_key



"""List images."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import image as image_mod
from SoftLayer import utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--name', default=None, help='Filter on image name')
@click.option('--public/--private', is_flag=True, default=None,
              help='Display only public or private images')
@click.option('--limit', '-l',
              help='How many results to get in one api call',
              default=100,
              show_default=True)
@environment.pass_env
def cli(env, name, public, limit):
    """List images."""

    image_mgr = SoftLayer.ImageManager(env.client)

    images = []
    if public in [False, None]:
        for image in image_mgr.list_private_images(name=name, limit=limit, mask=image_mod.MASK):
            images.append(image)

    if public in [True, None]:
        for image in image_mgr.list_public_images(name=name, limit=limit, mask=image_mod.MASK):
            images.append(image)

    table = formatting.Table(['Id', 'Name', 'Type', 'Visibility', 'Account', 'OS', 'Created', 'Notes'])
    table.align['OS'] = 'l'
    table.align['Notes'] = 'l'

    images = [image for image in images if not image['parentId']]
    for image in images:
        operative_system = '-'
        if image.get('children') and len(image.get('children')) != 0:
            if image.get('children')[0].get('blockDevices') and len(image.get('children')[0].get('blockDevices')) != 0:
                for block_device in image.get('children')[0].get('blockDevices'):
                    if block_device.get('diskImage').get('softwareReferences') and \
                            len(block_device.get('diskImage').get('softwareReferences')) != 0:
                        operative_system = block_device.get('diskImage').get('softwareReferences')[0].\
                            get('softwareDescription').get('longDescription')
        visibility = (image_mod.PUBLIC_TYPE if image['publicFlag'] else image_mod.PRIVATE_TYPE)
        table.add_row([
            image.get('id', formatting.blank()),
            formatting.FormattedItem(image['name'], click.wrap_text(image['name'], width=50)),
            formatting.FormattedItem(
                utils.lookup(image, 'imageType', 'keyName'),
                utils.lookup(image, 'imageType', 'name')),
            visibility,
            image.get('accountId', formatting.blank()),
            operative_system,
            utils.clean_time(image.get('createDate', formatting.blank())),
            image.get('note', formatting.blank()),
        ])

    env.fout(table)



getAllObjects = [
    {'id': 21, 'keyname': 'DEFAULT', 'name': 'Default'},
    {'id': 3, 'keyname': 'DNS', 'name': 'DNS'},
    {'id': 2, 'keyname': 'HTTP', 'name': 'HTTP'},
    {'id': 5, 'keyname': 'HTTP-CUSTOM', 'name': 'HTTP-CUSTOM'},
    {'id': 4, 'keyname': 'ICMP', 'name': 'Ping'},
    {'id': 1, 'keyname': 'TCP', 'name': 'TCP'}
]



"""
    SoftLayer.CLI.exceptions
    ~~~~~~~~~~~~~~~~~~~~~~~~
    Exceptions to be used in the CLI modules.

    :license: MIT, see LICENSE for more details.
"""


# pylint: disable=keyword-arg-before-vararg
class CLIHalt(SystemExit):
    """Smoothly halt the execution of the command. No error."""

    def __init__(self, code=0, *args):
        super().__init__(*args)
        self.code = code

    def __str__(self):
        message = getattr(self, 'message')
        return f"<CLIHalt code={self.code} msg={message}>"
    __repr__ = __str__


class CLIAbort(CLIHalt):
    """Halt the execution of the command. Gives an exit code of 2."""

    def __init__(self, msg, *args):
        super().__init__(code=2, *args)
        self.message = msg

    def __str__(self):
        message = getattr(self, 'message')
        return f"<CLIAbort code={self.code} msg={message}>"
    __repr__ = __str__


class ArgumentError(CLIAbort):
    """Halt the execution of the command because of invalid arguments."""

    def __init__(self, msg, *args):
        super().__init__(msg, *args)
        self.message = "Argument Error: %s" % msg

    def __str__(self):
        message = getattr(self, 'message')
        return f"<ArgumentError code={self.code} msg={message}>"
    __repr__ = __str__



"""List firewalls."""
# :license: MIT, see LICENSE for more details.

import os
import subprocess
import tempfile

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import firewall
from SoftLayer.CLI import formatting

DELIMITER = "=========================================\n"


def parse_rules(content=None):
    """Helper to parse the input from the user into a list of rules.

    :param string content: the content of the editor
    :returns: a list of rules
    """
    rules = content.split(DELIMITER)
    parsed_rules = []
    order = 1
    for rule in rules:
        if rule.strip() == '':
            continue
        parsed_rule = {}
        lines = rule.split("\n")
        parsed_rule['orderValue'] = order
        order += 1
        for line in lines:
            if line.strip() == '':
                continue
            key_value = line.strip().split(':')
            key = key_value[0].strip()
            value = key_value[1].strip()
            if key == 'action':
                parsed_rule['action'] = value
            elif key == 'protocol':
                parsed_rule['protocol'] = value
            elif key == 'source_ip_address':
                parsed_rule['sourceIpAddress'] = value
            elif key == 'source_ip_subnet_mask':
                parsed_rule['sourceIpSubnetMask'] = value
            elif key == 'destination_ip_address':
                parsed_rule['destinationIpAddress'] = value
            elif key == 'destination_ip_subnet_mask':
                parsed_rule['destinationIpSubnetMask'] = value
            elif key == 'destination_port_range_start':
                parsed_rule['destinationPortRangeStart'] = int(value)
            elif key == 'destination_port_range_end':
                parsed_rule['destinationPortRangeEnd'] = int(value)
            elif key == 'version':
                parsed_rule['version'] = int(value)
        parsed_rules.append(parsed_rule)
    return parsed_rules


def open_editor(rules=None, content=None):
    """Helper to open an editor for editing the firewall rules.

    This method takes two parameters, if content is provided,
    that means that submitting the rules failed and we are allowing
    the user to re-edit what they provided. If content is not provided, the
    rules retrieved from the firewall will be displayed to the user.

    :param list rules: A list containing the rules of the firewall
    :param string content: the content that the user provided in the editor
    :returns: a formatted string that get be pushed into the editor
    """

    # Let's get the default EDITOR of the environment,
    # use nano if none is specified
    editor = os.environ.get('EDITOR', 'nano')

    with tempfile.NamedTemporaryFile(suffix=".tmp") as tfile:

        if content:
            # if content is provided, just display it as is
            tfile.write(content)
            tfile.flush()
            subprocess.call([editor, tfile.name])
            tfile.seek(0)
            data = tfile.read()
            return data

        if not rules:
            # if the firewall has no rules, provide a template
            tfile.write(DELIMITER)
            tfile.write(get_formatted_rule())
        else:
            # if the firewall has rules, display those to the user
            for rule in rules:
                tfile.write(DELIMITER)
                tfile.write(get_formatted_rule(rule))
        tfile.write(DELIMITER)
        tfile.flush()
        subprocess.call([editor, tfile.name])
        tfile.seek(0)
        data = tfile.read()
        return data


def get_formatted_rule(rule=None):
    """Helper to format the rule into a user friendly format.

    :param dict rule: A dict containing one rule of the firewall
    :returns: a formatted string that get be pushed into the editor
    """
    rule = rule or {}
    return ('action: %s\n'
            'protocol: %s\n'
            'source_ip_address: %s\n'
            'source_ip_subnet_mask: %s\n'
            'destination_ip_address: %s\n'
            'destination_ip_subnet_mask: %s\n'
            'destination_port_range_start: %s\n'
            'destination_port_range_end: %s\n'
            'version: %s\n'
            % (rule.get('action', 'permit'),
               rule.get('protocol', 'tcp'),
               rule.get('sourceIpAddress', 'any'),
               rule.get('sourceIpSubnetMask', '255.255.255.255'),
               rule.get('destinationIpAddress', 'any'),
               rule.get('destinationIpSubnetMask', '255.255.255.255'),
               rule.get('destinationPortRangeStart', 1),
               rule.get('destinationPortRangeEnd', 1),
               rule.get('version', 4)))


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Edit firewall rules."""

    mgr = SoftLayer.FirewallManager(env.client)

    firewall_type, firewall_id = firewall.parse_id(identifier)
    if firewall_type == 'vlan':
        orig_rules = mgr.get_dedicated_fwl_rules(firewall_id)
    else:
        orig_rules = mgr.get_standard_fwl_rules(firewall_id)
    # open an editor for the user to enter their rules
    edited_rules = open_editor(rules=orig_rules)
    env.out(edited_rules)
    if formatting.confirm("Would you like to submit the rules. "
                          "Continue?"):
        while True:
            try:
                rules = parse_rules(edited_rules)
                if firewall_type == 'vlan':
                    mgr.edit_dedicated_fwl_rules(firewall_id, rules)
                else:
                    mgr.edit_standard_fwl_rules(firewall_id, rules)
                break
            except (SoftLayer.SoftLayerError, ValueError) as error:
                env.out("Unexpected error({%s})" % (error))
                if formatting.confirm("Would you like to continue editing "
                                      "the rules. Continue?"):
                    edited_rules = open_editor(content=edited_rules)
                    env.out(edited_rules)
                    if formatting.confirm("Would you like to submit the "
                                          "rules. Continue?"):
                        continue
                    raise exceptions.CLIAbort('Aborted.')
                raise exceptions.CLIAbort('Aborted.')
        env.fout('Firewall updated!')
    else:
        raise exceptions.CLIAbort('Aborted.')



"""List existing replicant volumes for a block volume."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import columns as column_helper
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import storage_utils

COLUMNS = storage_utils.REPLICATION_PARTNER_COLUMNS
DEFAULT_COLUMNS = storage_utils.REPLICATION_PARTNER_DEFAULT


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('volume-id')
@click.option('--columns',
              callback=column_helper.get_formatter(COLUMNS),
              help=f"Columns to display. Options: {', '.join(column.name for column in COLUMNS)}",
              default=','.join(DEFAULT_COLUMNS))
@click.option('--sortby', help='Column to sort by', default='Username')
@environment.pass_env
def cli(env, columns, sortby, volume_id):
    """List existing replica volumes for a block volume."""
    block_storage_manager = SoftLayer.BlockStorageManager(env.client)

    legal_volumes = block_storage_manager.get_replication_partners(
        volume_id
    )

    if not legal_volumes:
        click.echo("There are no replication partners for the given volume.")
    else:
        table = formatting.Table(columns.columns)
        table.sortby = sortby
        for legal_volume in legal_volumes:
            table.add_row([value or formatting.blank()
                           for value in columns.row(legal_volume)])

        env.fout(table)



"""
    SoftLayer.search
    ~~~~~~~~~~~~~~~~
    Search Manager

    :license: MIT, see LICENSE for more details.
"""


class SearchManager(object):
    """Manager to help search via the SoftLayer API.

    :param SoftLayer.API.BaseClient client: the client instance
    """

    def __init__(self, client):
        self.client = client
        self.search_manager = client['SoftLayer_Search']

    def get_object_types(self):
        """returns a collection of SoftLayer_Container_Search_ObjectType containers."""
        return self.search_manager.getObjectTypes()

    def search(self, search_string):
        """allows for searching for SoftLayer resources by simple phrase."""
        return self.search_manager.search(search_string)

    def advanced(self, search_string):
        """Uses the SoftLayer_Search::advancedSearch API. Allows for more complicated search phrases."""
        return self.search_manager.advancedSearch(search_string)

    def search_instances(self, search_string, mask=None, **kwargs):
        """Lists VSIs based in the search_string.

        Also takes in a few search terms as \\*\\*kwargs. such as hostname, datacenter, domain and tags
        """

        # This forces the Search API to do a fuzzy search on our term, kinda. Not sure why the ** are
        # Required but it will do an exact search without them.
        if search_string:
            search_string = f"*{search_string}*"
        search_string = f"_objectType:SoftLayer_Virtual_Guest {search_string}"
        if kwargs.get('hostname'):
            search_string = f"{search_string} hostname: *{kwargs.get('hostname')}*"
        if kwargs.get('domain'):
            search_string = f"{search_string} domain: *{kwargs.get('domain')}*"
        if kwargs.get('datacenter'):
            search_string = f"{search_string} datacenter.longName: *{kwargs.get('datacenter')}*"
        if kwargs.get('tags'):
            tags = " ".join(f"tagReferences.tag.name: \"{t}\"" for t in kwargs.get("tags", []))
            search_string = f"{search_string} {tags}"
        result = self.search_manager.advancedSearch(search_string, mask=mask)
        guests = []
        for resource in result:
            guests.append(resource.get('resource'))

        return guests

    def search_hadrware_instances(self, search_string, mask=None, **kwargs):
        """Lists hardwares based in the search_string.

        Also takes in a few search terms as \\*\\*kwargs. such as hostname, datacenter, domain and tags
        """

        # This forces the Search API to do a fuzzy search on our term, kinda. Not sure why the ** are
        # Required but it will do an exact search without them.
        if search_string:
            search_string = f"*{search_string}*"
        search_string = f"_objectType:SoftLayer_Hardware {search_string}"
        if kwargs.get('hostname'):
            search_string = f"{search_string} hostname: *{kwargs.get('hostname')}*"
        if kwargs.get('domain'):
            search_string = f"{search_string} domain: *{kwargs.get('domain')}*"
        if kwargs.get('cpu'):
            search_string = f"{search_string} cpu: *{kwargs.get('cpu')}*"
        if kwargs.get('network'):
            search_string = f"{search_string} network: *{kwargs.get('network')}*"
        if kwargs.get('memory'):
            search_string = f"{search_string} memory: *{kwargs.get('memory')}*"
        if kwargs.get('datacenter'):
            search_string = f"{search_string} datacenter.longName: *{kwargs.get('datacenter')}*"
        if kwargs.get('tags'):
            tags = " ".join(kwargs.get("tags", []))
            search_string = f"{search_string} internalTagReferences.tag.name: {tags}"
        result = self.search_manager.advancedSearch(search_string, mask=mask)
        servers = []
        for resource in result:
            servers.append(resource.get('resource'))

        return servers



"""Firewalls."""
# :license: MIT, see LICENSE for more details.

from SoftLayer.CLI import exceptions


def parse_id(input_id):
    """Helper package to retrieve the actual IDs.

    :param input_id: the ID provided by the user
    :returns: A list of valid IDs
    """
    key_value = input_id.split(':')

    if len(key_value) != 2:
        raise exceptions.CLIAbort(
            'Invalid ID %s: ID should be of the form xxx:yyy' % input_id)
    return key_value[0], int(key_value[1])



IMAGES = [{
    'accountId': 1234,
    'blockDevices': [],
    'createDate': '2013-12-05T21:53:03-06:00',
    'globalIdentifier': '0B5DEAF4-643D-46CA-A695-CECBE8832C9D',
    'id': 100,
    'name': 'test_image',
    'parentId': '',
    'publicFlag': True,
    'children': [{
        'datacenter': {
            'name': 'ams01'
        }
    }],
}, {
    'accountId': 1234,
    'blockDevices': [],
    'createDate': '2013-12-05T21:53:03-06:00',
    'globalIdentifier': 'EB38414C-2AB3-47F3-BBBD-56A5F689620B',
    'id': 101,
    'name': 'test_image2',
    'parentId': '',
    'publicFlag': True,
    'children': [{
        'datacenter': {
            'name': 'ams01'
        }
    }],
}]

getObject = IMAGES[0]
getPublicImages = IMAGES
deleteObject = {}
editObject = True
setTags = True
createFromExternalSource = {
    'createDate': '2013-12-05T21:53:03-06:00',
    'globalIdentifier': '0B5DEAF4-643D-46CA-A695-CECBE8832C9D',
    'id': 100,
    'name': 'test_image',
}
createFromIcos = {
    'createDate': '2013-12-05T21:53:03-06:00',
    'globalIdentifier': '0B5DEAF4-643D-46CA-A695-CECBE8832C9D',
    'id': 100,
    'name': 'test_image',
}
copyToExternalSource = True
copyToIcos = True
addLocations = True
removeLocations = True
getStorageLocations = [
    {'id': 265592, 'longName': 'Amsterdam 1', 'name': 'ams01', 'statusId': 2},
    {'id': 814994, 'longName': 'Amsterdam 3', 'name': 'ams03', 'statusId': 2},
]
permitSharingAccess = True
denySharingAccess = True
